--[[   
   vectorial3.lua ver 0.2 - A library for 3D vectors.
   Copyright (C) 2015 Leo Tindall
   ---
    All operators work as expected except modulo (%) which is vector distance and concat (..) which is linear distance.
   ---
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 3 of the License, or
   (at your option) any later version.
   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.
   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software Foundation,
   Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301  USA
]]
local module = function (ix, iy, iz)
	local v3d = {}
		v3d.type = "vec"
	      v3d.X = ix or 0
	      v3d.Y = iy or 0
	      v3d.Z = iz or 0
	      mt = {} --Metatable
	
	function v3d:deepcopy(orig) --Deeply copy a table. This is for the operation metatables.
		local orig_type = type(orig)
		local copy
		if orig_type == 'table' then
			copy = {}
			for orig_key, orig_value in next, orig, nil do
					copy[self:deepcopy(orig_key)] = self:deepcopy(orig_value)
			end
			setmetatable(copy, self:deepcopy(getmetatable(orig)))
   		else -- number, string, boolean, etc
			copy = orig
    	end
    	return copy
	end
	
	--Vector Specific Math

	function v3d:getAngle() --Return the 3D angle (heading, carom) of the vector IN RADIANS!.
		local hdg = math.atan2(self:getY(), self:getX())
		local crm = math.atan2(self:getZ(), 0)
		return hdg, crm
	end
	
	function v3d:getLength()
		return math.sqrt( self:x()^2 + self:y()^2 + self:z()^2 )
	end

	--Comparisons

	mt.__eq = function(lhs, rhs)
		--Equal To operator for Vector3Ds
		return (lhs:getX() == rhs:getX()) and (lhs:getY() == rhs:getY()) and (lhs:getZ() == rhs:getZ())
	end
	
	mt.__lt = function(lhs, rhs)
		--Less Than operator for Vector3Ds
		print(lhs, rhs)
		return lhs:getLength() < rhs:getLength() --We do this to compute the linear value of the vector so that, for example, (a % b) < (c % d) will not be broken.
	end
	
	mt.__le = function(lhs, rhs)
		--Less Than Or Equal To operator for Vector3Ds
		return lhs:getLength() <= rhs:getLength() --We do this to compute the linear value of the vector so that, for example, (a % b) < (c % d) will not be broken.
	end
	
	--Operations
	
	function v3d:setX(x)
		self.X = x
	end
	
	function v3d:setY(y)
		self.Y = y
	end

	function v3d:setZ(z)
		self.Z = z
	end
	
	function v3d:x()
		return self.X
	end
	
	function v3d:y()
		return self.Y
	end

	function v3d:z()
		return self.Z
	end
	
	function v3d:getX()
		return self.X
	end
	
	function v3d:getY()
		return self.Y
	end

	function v3d:getZ()
		return self.Z
	end
	
	function v3d:PolyNormal( v2, v3 )
		return ( self - v2 )^( self - v3 )
	end
	
	function v3d:normal()
		local l = self:getLength()
		local out = self:deepcopy(self)
		out:setX( self:x()/l )
		out:setY( self:y()/l )
		out:setZ( self:z()/l )
		return out
	end
	
	function v3d:dot( p2 )
		return self:x()*p2:x() + self:y()*p2:y() + self:z()*p2:z()
	end
	
	mt.__unm = function(rhs)
		--Unary Minus (negation) operator for Vector3Ds
		out = rhs:deepcopy(rhs) --Copy the operand for the output (else the output won't have metamethods)
		out:setX(-rhs:getX()) --Operate on the X property
		out:setY(-rhs:getY()) --Operate on the Y property
		out:setY(-rhs:getZ()) --Operate on the Z property
		return out
	end
	
	mt.__add = function(lhs, rhs)
		--Addition operator for Vector3Ds
		out = lhs:deepcopy(lhs)--Copy the operand for the output (else the output won't have metamethods)
		out:setX(lhs:getX() + rhs:getX()) --Operate on the X property
		out:setY(lhs:getY() + rhs:getY()) --Operate on the Y property
		out:setZ(lhs:getZ() + rhs:getZ()) --Operate on the Z property
		return out
	end

	mt.__sub = function(lhs, rhs)
		--Subtraction operator for Vector3Ds
		out = lhs:deepcopy(lhs)--Copy the operand for the output (else the output won't have metamethods)
		out:setX(lhs:getX() - rhs:getX()) --Operate on the X property
		out:setY(lhs:getY() - rhs:getY()) --Operate on the Y property
		out:setZ(lhs:getZ() - rhs:getZ()) --Operate on the Z property
		return out
	end

	mt.__mul = function(lhs, rhs)
		--Multiplication operator for Vector3Ds
		out = lhs:deepcopy(lhs)--Copy the operand for the output (else the output won't have metamethods)
		out:setX(lhs:getX() * rhs:getX()) --Operate on the X property
		out:setY(lhs:getY() * rhs:getY()) --Operate on the Y property
		out:setZ(lhs:getZ() * rhs:getZ()) --Operate on the Z property
		return out
	end

	mt.__div = function(lhs, rhs)
		--Division operator for Vector3Ds
		out = lhs:deepcopy(lhs)--Copy the operand for the output (else the output won't have metamethods)
		out:setX(lhs:getX() / rhs:getX()) --Operate on the X property
		out:setY(lhs:getY() / rhs:getY()) --Operate on the Y property
		out:setZ(lhs:getZ() / rhs:getZ()) --Operate on the Z property
		return out
	end
	
	mt.__mod = function(lhs, rhs)
		--Vector distance operator for Vector3Ds. Denoted by modulo (%)
		out = lhs:deepcopy(lhs)		--Copy the operand for the output (else the output won't have metamethods)
		out:setX(math.abs(rhs:getX() - lhs:getX())) --Operate on the X property
		out:setY(math.abs(rhs:getY() - lhs:getY())) --Operate on the Y property
		out:setZ(math.abs(rhs:getZ() - lhs:getZ())) --Operate on the Z property
		return out	
	end

	mt.__concat = function(lhs, rhs)
		--Linear distance operator for Vector3Ds. Denoted by concat (..)
		out = 0		--This is a linear operation, so no deepcopy. 
		out = math.sqrt(((rhs:getX() - lhs:getX())^2) + (((rhs:getY() - lhs:getY())^2) + ((rhs:getZ() - lhs:getZ())^2))) --Distance formula
		return out
	end

	mt.__tostring = function(s)
		--tostring handler for Vector3D
		out = ""	--This is a string operation, so no deepcopy.
		out = "[(X:"
		out = out .. s:getX() 
		out = out .. "),(Y:" 
		out = out .. s:getY() 
		out = out .. "),(Z:" 
		out = out .. s:getZ() 
		out = out .. ")]"
		return out
	end
	
	-- cross
	function mt.__pow(p1,p2)
		local out = p1:deepcopy(p1)	
	   
		out:setX( p1:y()*p2:z() - p1:z()*p2:y() )
		out:setY( p1:z()*p2:x() - p1:x()*p2:z() )
		out:setZ( p1:x()*p2:y() - p1:y()*p2:x() )
		return out

	end
	
	setmetatable(v3d, mt)

	return v3d
end
return module